package name.panitz.longStack;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.List;
import java.util.Arrays;
import java.util.Collection;

import org.junit.Before;
import org.junit.Test;

import static name.panitz.longStack.AST.*;
import static name.panitz.longStack.LongStackParser.*;

public class ASTTest {
  int i=0;
  List<Fun> funs = List.of(factorialRek,factorial,minus,fib);
  @Before
  public void setUp() throws Exception {
    i=0;
  }

  @Test
  public void evalFac5() {
    assertEquals("",120,new FunCall("fac", List.of(new IntLiteral(5))).ev(funs));
  }
  @Test
  public void evalF5() {
    assertEquals("",120,new FunCall("f", List.of(new IntLiteral(5))).ev(funs));
  }
  @Test
  public void evalFib10() {
    assertEquals("",55,new FunCall("fib", List.of(new IntLiteral(10))).ev(funs));
  }
  @Test
  public void evalMinus() {
    assertEquals("",55,new FunCall("minus", List.of(new IntLiteral(65),new IntLiteral(10))).ev(funs));
  }

  
  @Test
  public void eval1add1() {
    assertEquals("",2,LongStackParser.parseExpr("1+1").ev());
  }
  @Test
  public void eval17add4Mul2_25() {
    assertEquals("",25,LongStackParser.parseExpr("17+4*2").ev());
  }
  @Test
  public void eval17add4Mul2_42() {
    assertEquals("",42,LongStackParser.parseExpr("(17+4)*2").ev());
  }

    
  @Test
  public void show42() {
    assertEquals("","42",LongStackParser.parseExpr("42").show());
  }
  @Test
  public void show2Add2() {
    assertEquals("","2+2",LongStackParser.parseExpr(" 2 + 2 ").show().replaceAll("\\s",""));
  }
  @Test
  public void roundtrip2Add2_25() {
    roundtrip("17+4*2");
  }
  @Test
  public void evalShow_42_1() {
    assertEquals("",42,evalShow("(17+4)*2"));
  }
  @Test
  public void evalShow_42_2() {
    assertEquals("",42,evalShow("(1+1)*(17+4)"));
  }
  @Test
  public void roundtripif1() {
    roundtrip("if x=42 then y:= a else 9");
  }
  @Test
  public void roundtripFacIt() {
    roundtrip(facIt);
  }

  @Test
  public void asm1Add1() {
  assertEquals("","movq $1, %rax\npushq %rax\nmovq $1, %rax\npopq %rbx\naddq %rbx, %rax".replaceAll("\\s",""),genCode("1+1").replaceAll("\\s",""));
  }

  @Test
  public void asm17add4mult2() {
    assertEquals("","movq$2,%raxpushq%raxmovq$4,%raxpopq%rbximulq%rbx,%raxpushq%raxmovq$17,%raxpopq%rbxaddq%rbx,%rax",genCode("17+4*2").replaceAll("\\s",""));
  }

  
  static String facIt = "{x:=5;r:=1;while x do {r:=x*r;x:=x-1};r} ";

  static String genCode(String s){
    var ast1 = parseExpr(s);
    return ast1.asm();
  }

  static long evalShow(String s){
    var ast1 = parseExpr(s);
    var s1 = ast1.show();
    try{
      var ast2 = parseExpr(s1);
      return ast2.ev();
    }catch (Exception e){
      fail("cannot parse result of show for: "+s+" unparsable source:" +s1);
      return 0L;
    }
  }

  
  
  static void roundtrip(String s){
    var ast1 = parseExpr(s);
    var s1 = ast1.show();
    try{
      var ast2 = parseExpr(s1);
      var s2 = ast2.show();
      assertEquals("roundtrip show -> parse -> does not yield same source code for: "+s,s1,s2);
    }catch (Exception e){
      fail("cannot parse result of show for: "+s+" unparsable source:" +s1);
    }
  }
}